'''nxs-analysis-tools package metadata.'''

# keep consistent with pyproject.toml
__project__ = 'nxs-analysis-tools'
__author__ = 'Steven J. Gomez Alvarado'
__email__ = 'stevenjgomez@ucsb.edu'
__copyright__ = f"2023, {__author__}"
__license__ = 'MIT'
__version__ = '0.0.39'
__repo_url__ = 'https://github.com/stevenjgomez/nxs_analysis_tools'
